/*
  Large integer division algorithm.

  04/2019 : F. Prill, DWD.

  The division algorithm is explained in

  P. B. Hansen: Multiple-Length Division Revisited: A Tour of the Minefield (192).
                Electrical Engineering and Computer Science Technical Reports. 166.
*/
#include "int_div.h"
#include <stdio.h>
#include <inttypes.h>

t_lid_int const bb[4] = { 1, 1200, 1440000, 1728000000 }; /* powers of b = 1200 */


/* initialize digits m,m+1,...,W_DIGITS by a constant val. */
void number_init(t_large_int *x, const t_lid_int val, const unsigned int m) {
  int i;
  for (i=m; i<=W_DIGITS; ++i)  (*x)[i]=val;
}


/* linear search to determine the length of a natural number. */
unsigned int length(const t_large_int x) {
  int i = W_DIGITS;
  while ((i >= 0) && (x[i] == iZero))  i--;
  return (i+1);
}


/* partial arithmetic wrt. radix b: product x=y*k, return ret!=0 if overflow. */
void product(t_large_int *x, const t_large_int y, const t_lid_int k, const t_lid_int b, int* ret) {
  int   i, m=length(y);
  t_lid_int temp, carry = 0;

  if (*ret != 0) return;
  for (i=0; i<m; ++i) {
    temp = y[i]*k + carry;
    (*x)[i] = temp % b;
    carry = temp/b;
  }
  if (m<=W_DIGITS)
    (*x)[m] = carry;
  else
    *ret = (carry != iZero);
  for (i=m+1; i<=W_DIGITS; ++i)  (*x)[i] = iZero;
}


/* partial arithmetic wrt. radix b: quotient x=y/k. */
void quotient(t_large_int *x, const t_large_int y, t_lid_int k, const t_lid_int b) {
  int i, m = length(y);
  t_lid_int temp, carry = iZero;
  number_init(x, iZero, m);

  for (i=m-1; i>=0; i--) {
    temp = carry*b + y[i];
    (*x)[i] = temp / k;
    carry = temp % k;
  }
}


/* partial arithmetic wrt. radix b: remainder x of y/k. */
void iremainder(t_large_int *x, const t_large_int y, t_lid_int k, const t_lid_int b) {
  int m = length(y), i;
  t_lid_int carry = iZero;

  number_init(x, iZero, 1);
  for (i=m-1; i>=0; i--)
    carry = (carry*b + y[i]) % k;
  (*x)[0] = carry;
}


/* prefix arithmetic wrt. radix b: estimate of a quotient digit, return <0 if error. */
t_lid_int trial(const t_large_int r, const t_large_int d,
		const int k, const int m, const t_lid_int b, int* ret) {
  t_lid_int d2,r3,t = iZero;
  int   km = k+m;
#if DEBUG
  if ((m < 2) || (km < m) || (km>W_DIGITS))  *ret = Fail;
#endif
  if (*ret != 0) return iZero;
  r3 = (r[km]*b + r[km-1])*b + r[km-2];
  d2 = d[m-1]*b + d[m-2];

  t = r3/d2;
  return (t > b-1) ? (b-1) : t;
}


/* prefix arithmetic: prefix comparison r{m+1} < dq, return <0 if error. */
int prefix_less_than(const t_large_int r, const t_large_int dq, const int k, const int m) {
  int i=m;

#if DEBUG
  if ((k < 0) || (k+m < k) || (k+m>W_DIGITS))  return Fail;
#endif
  while ((r[i+k] == dq[i]) && (i>0))  i--;
  return r[i+k] < dq[i];
}


/* prefix arithmetic wrt. radix b: prefix subtraction r{m+1} := r{m+1} - dq, return ret!=0 if error. */
void difference(t_large_int *r, const t_large_int dq, const int k, const int m, const t_lid_int b, int *ret) {
  int i;
  t_lid_int borrow, diff;

#ifdef DEBUG
  if ((k < 0) || (k+m < k) || (k+m>W_DIGITS))  *ret = Fail;
#endif
  if (*ret != 0) return;
  borrow = iZero;
  for (i=0; i<=m; i++) {
    diff = (*r)[i+k] - dq[i] - borrow + b;
    (*r)[i+k] = diff % b;
    borrow = (t_lid_int) 1 - diff / b;
  }
  *ret = (borrow != iZero);
}


/* complete algorithm for multiple length division x/y wrt. radix b, return <0 if error. */
int division(const t_large_int x, const t_large_int y,
             t_large_int *q, t_large_int *r, const t_lid_int b) {
  int i, k, m, n, ret = 0;
  t_lid_int y1, f, qt;
  t_large_int d, dq, rm;

  m = length(y);
  if (m == 1) {
    y1 = y[m-1];
    if (y1 <= 0) return Fail;
    quotient(q,x,y1,b);
    iremainder(r,x,y1,b);
  }
  else {
    n = length(x);
    number_init(q, iZero, 0);
    number_init(r, iZero, 0);
    if (m > n)
      for (i=0; i<n; i++)  (*r)[i] = x[i];
    else {
      /* long division algorithm */
      if ((m < 2) || (n < m) || (n>W_DIGITS)) return Fail;
      f = b / (y[m-1] + (t_lid_int) 1);
      product(&rm,x,f,b, &ret);
      product(&d, y,f,b, &ret);
      for (k=n-m; (k>=0) && (ret == 0); k--) {
        qt = trial(rm, d, k, m, b, &ret);
        product(&dq,d,qt,b, &ret);

        if ((ret == 0) && (prefix_less_than(rm,dq,k,m)))
          product(&dq,d,--qt,b, &ret);

        (*q)[k] = qt;
        difference(&rm,dq,k,m,b, &ret);
      }
      quotient(r,rm,f,b);
    }
  }
  return ret;
}


/* convert to basis "b" by a succession of Euclidean divisions, stores to v_out[istart...]. */
void convert_to_basis(const t_lid_int v_in, t_large_int *v_out,
		      const int digits, const int istart) {
  int i;
  t_lid_int v = v_in;
  for (i=digits; i>=0; i--) {
    (*v_out)[istart+i] = v / bb[i];
    v %= bb[i];
  }
}


/* convert to decimal basis. */
t_lid_int convert_to_decimal(const t_large_int v_in,
			     const int istart, const int iend) {
  int i, l = length(v_in);
  t_lid_int v_out = iZero;

  if (iend < l)  l=iend;
  if ((l-istart) > 3)  return (t_lid_int) Fail;
  for (i=0; i<(l-istart+1); i++)
    v_out += v_in[i+istart] * bb[i];
  return v_out;
}


/* print number to screen. */
void  print_large_int(const t_large_int x) {
  int i, m = length(x) - 1;
  for (i=m; i>=0; i--)
    printf("%" PRId64 " ", x[i]);
  printf("\n");
}

