/*
  Large integer division algorithm for (days + msecs).

  04/2019 : F. Prill, DWD.

  The method uses a basis b=1200 representation for time spans and the
  division algorithm explained in

  P. B. Hansen: Multiple-Length Division Revisited: A Tour of the Minefield (192).
                Electrical Engineering and Computer Science Technical Reports. 166.
*/
#include "timespan_div.h"
#include <stdio.h>

const t_lid_int msecs_day = 86400000;

/*
   convert days and day fraction to b=1200 basis by a succession of
   Euclidean divisions.
   By pre-multiplying with a factor of 20, the day fraction can be
   expressed by three digits wrt. the basis b=1200.

   restrictions:

   - the no. of digits in the algorithm restricts numbers to 1200^3
     days, ie. approx. 4.7 mio years.
   - the quotient is limited to 1200^3 in order to fit into int64_t
     after conversion to decimal base.
   - all numbers and time spans must be positive.
*/
int divide_timespan(const t_lid_int days1, const t_lid_int day_fraction1,
                    const t_lid_int days2, const t_lid_int day_fraction2,
                    t_lid_int *q_decimal, t_lid_int *remainder_days, t_lid_int *remainder_ms)
{
  int      ret = 0;
  t_large_int p1, p2, q, r; /* dividend, divisor, quotient, remainder */
  t_lid_int    qd = 0, rd = 0;

  if ((day_fraction1 > msecs_day) || (day_fraction2 > msecs_day))  printf("Fail 1\n");
  if ((day_fraction1 > msecs_day) || (day_fraction2 > msecs_day))  return Fail;

  convert_to_basis(day_fraction1 * (t_lid_int) 20, &p1, 2, 0);
  convert_to_basis(days1, &p1, W_DIGITS-3, 3);
  convert_to_basis(day_fraction2 * (t_lid_int) 20, &p2, 2, 0);
  convert_to_basis(days2, &p2, W_DIGITS-3, 3);

  ret = division(p1, p2, &q, &r, (t_lid_int) 1200);  /* perform division. */

  if (ret == 0) {
    /* convert result back to b=10 basis. */
    *q_decimal      = qd = convert_to_decimal(q,0,W_DIGITS);
    *remainder_days =      convert_to_decimal(r,3,W_DIGITS);
    *remainder_ms   = rd = convert_to_decimal(r,0,2);
    ret = (qd < iZero) || (rd < iZero);

    *remainder_ms /= ((t_lid_int) 20);
  }
  return ret;
}


